from .manager import Manager
from .query import QuerySet
from .base import Model
from .pager import Pager, Page
from .cenum import CEnum, CREnum
from .fields import *  # NOQA
from .fields import __all__ as fields_all

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import signals
from django.db.models.aggregates import *  # NOQA
from django.db.models.aggregates import __all__ as aggregates_all
from django.db.models.constraints import *  # NOQA
from django.db.models.constraints import __all__ as constraints_all
from django.db.models.deletion import (
    CASCADE, DO_NOTHING, PROTECT, SET, SET_DEFAULT, SET_NULL, ProtectedError,
)
from django.db.models.expressions import (
    Case, Exists, Expression, ExpressionList, ExpressionWrapper, F, Func,
    OuterRef, RowRange, Subquery, Value, ValueRange, When, Window, WindowFrame,
)
from django.db.models.fields.files import FileField, ImageField
from django.db.models.fields.proxy import OrderWrt
from django.db.models.indexes import *  # NOQA
from django.db.models.indexes import __all__ as indexes_all
from django.db.models.lookups import Lookup, Transform
from django.db.models.query import (
    Prefetch, Q, prefetch_related_objects,
)
from django.db.models.query_utils import FilteredRelation

# Imports that would create circular imports if sorted
from django.db.models.base import DEFERRED  # isort:skip
from django.db.models.fields.related import (  # isort:skip
    ForeignKey, ForeignObject, OneToOneField, ManyToManyField,
    ManyToOneRel, ManyToManyRel, OneToOneRel,
)


__all__ = aggregates_all + constraints_all + fields_all + indexes_all
__all__ += [
    ObjectDoesNotExist, signals,
    CASCADE, DO_NOTHING, PROTECT, SET, SET_DEFAULT, SET_NULL,
    ProtectedError,
    Case, Exists, Expression, ExpressionList, ExpressionWrapper, F,
    Func, OuterRef, RowRange, Subquery, Value, ValueRange, When,
    Window, WindowFrame,
    FileField, ImageField, OrderWrt, Lookup, Transform, Manager,
    Prefetch, Q, QuerySet, prefetch_related_objects, DEFERRED, Model,
    FilteredRelation,
    ForeignKey, ForeignObject, OneToOneField, ManyToManyField,
    ManyToOneRel, ManyToManyRel, OneToOneRel,
]



class Classify:
    def dict(self, *args):
        if args:
            dict_ = dict()
            for k in args:
                dict_[k] = self._dict.get(k)
            return dict_
        return self._dict

    def __init__(self, d):
        if not isinstance(d, dict):
            return
        self._dict = d

    def __getattr__(self, item):
        return self._dict.get(item)



from django.db import models

from .query import QuerySet


class Manager(models.Manager):
    def __init__(self):
        super().__init__()
        self.restrict_args = tuple()
        self.restrict_kwargs = dict()

    def get_queryset(self):
        return QuerySet(self.model, using=self._db).filter(*self.restrict_args,
                                                           **self.restrict_kwargs)

    def restrict(self, *args, **kwargs):
        self.restrict_args = args
        self.restrict_kwargs = kwargs
        return self

    def search(self, *args, **kwargs):
        return self.all().search(*args, **kwargs)

    def dict(self, dictor):
        return self.all().dict(dictor)

    def page(self, pager, last=0, count=5):
        return pager.page(self.all(), last=last, count=count)



import datetime
import warnings
from typing import List, Tuple, Optional

from ..error import E
from ..http_code import HttpCode as Hc
from django.db import models
from smartify import Attribute, P, PError

from .manager import Manager
from . import fields


@E.register()
class ModelError:
    FIELD_FORMAT = E("Field format error", hc=Hc.Forbidden)


class Constraint:
    LENGTH_T = "{0}({1})'s length should not %s{2}"
    VALUE_T = "{0}({1}) should not %s{2}"

    def __init__(self, field, type_, boundary=True, compare=None, template=None):
        self.field, self.type, self.boundary = field, type_, boundary
        self.compare = compare or (lambda x: x)
        self.error_template = template or self.VALUE_T

    def fit(self, field, value):
        if isinstance(field, fields.CharField):
            max_, min_ = field.max_length, field.min_length
        else:
            max_, min_ = field.max_value, field.min_value
        if max_ and max_ < self.compare(value):
            raise ModelError.FIELD_FORMAT(append_message=(self.error_template % 'bigger').format(
                field.name, field.verbose_name, max_))
        if min_ and min_ > self.compare(value):
            raise ModelError.FIELD_FORMAT(append_message=(self.error_template % 'smaller').format(
                field.name, field.verbose_name, min_))


CONSTRAINTS = [
    Constraint(fields.CharField, str, compare=lambda x: len(x), template=Constraint.LENGTH_T),
    Constraint(fields.IntegerField, int),
    Constraint(fields.FloatField, float),
    Constraint(fields.BooleanField, bool, boundary=False),
    Constraint(fields.DateField, datetime.date),
    Constraint(fields.DateTimeField, datetime.datetime),
]


class Model(models.Model):
    objects = Manager()

    class Meta:
        abstract = True
        default_manager_name = 'objects'

    @classmethod
    def get_fields(cls, *field_names: str) -> Tuple[models.Field]:
        field_jar = dict()
        for field in cls._meta.fields:
            field_jar[field.name] = field

        field_list = []  # type: List[models.Field]
        for field_name in field_names:
            field_list.append(field_jar.get(field_name))

        return tuple(field_list)

    @classmethod
    def get_field(cls, field_name) -> models.Field:
        return cls.get_fields(field_name)[0]

    @classmethod
    def get_params(cls, *field_names: str):
        return tuple(map(cls.get_param, field_names))

    @classmethod
    def get_param(cls, field_name):
        field = cls.get_field(field_name)
        p = P(field.name, read_name=field.verbose_name)
        p.allow_null = field.null
        p.validate(Model.field_validator(field))
        return p

    P = get_params

    @staticmethod
    def field_validator(field: models.Field):
        attr = field.name
        verbose = field.verbose_name
        cls = field.model

        def validate(value):
            for constraint in CONSTRAINTS:
                if isinstance(field, constraint.field):
                    if not isinstance(value, constraint.type):
                        raise ModelError.FIELD_FORMAT(
                            append_message='%s(%s) has wrong type' % (attr, verbose))
                    if constraint.boundary:
                        constraint.fit(field, value)
                    break

            if field.choices:
                choice_match = False
                for choice in field.choices:
                    if value == choice[0]:
                        choice_match = True
                        break
                if not choice_match:
                    raise ModelError.FIELD_FORMAT(
                        append_message='%s(%s) is beyond choices' % (attr, verbose))

            custom_validator = getattr(cls, '_valid_%s' % attr, None)
            if callable(custom_validator):
                try:
                    custom_validator(value)
                except E as e:
                    raise e
                except Exception as err:
                    raise PError.VALIDATOR_CRUSHED(attr, verbose, debug_message=str(err))

        return validate

    @classmethod
    def validator(cls, attr_jar: dict):
        field_dict = dict()
        for field in cls._meta.fields:
            field_dict[field.name] = field

        for attr in attr_jar:
            attr_value = attr_jar[attr]
            if attr in field_dict:
                field = field_dict[attr]  # type: Optional[models.Field]
            else:
                field = None

            if field:
                try:
                    cls.field_validator(field)(attr_value)
                except E as e:
                    raise e
                except Exception as err:
                    raise PError.VALIDATOR_CRUSHED(attr, field.verbose_name, debug_message=str(err))
            else:
                custom_validator = getattr(cls, '_valid_%s' % attr, None)
                if callable(custom_validator):
                    try:
                        custom_validator(attr_value)
                    except E as e:
                        raise e
                    except Exception as err:
                        raise PError.VALIDATOR_CRUSHED(attr, attr, debug_message=str(err))

    def dictor(self, *field_list):
        warnings.warn(
            'dictor method is deprecated, use dictify instead.',
            DeprecationWarning)
        return self.dictify(*field_list)

    def dictify(self, *field_list):
        return Attribute.dictify(self, *field_list)



from django.db.models.fields import *


class CharField(CharField):
    def __init__(self, *, min_length=None, **kwargs):
        self.min_length = min_length
        super().__init__(**kwargs)


class FloatField(FloatField):
    def __init__(self, *, max_value=None, min_value=None, **kwargs):
        self.max_value, self.min_value = max_value, min_value
        super().__init__(**kwargs)


class IntegerField(IntegerField):
    def __init__(self, *, max_value=None, min_value=None, **kwargs):
        self.max_value, self.min_value = max_value, min_value
        super().__init__(**kwargs)


class PositiveIntegerField(PositiveIntegerField, IntegerField):
    pass


class PositiveSmallIntegerField(PositiveSmallIntegerField, IntegerField):
    pass


class SmallIntegerField(SmallIntegerField, IntegerField):
    pass


class BigIntegerField(BigIntegerField, IntegerField):
    pass


class DateField(DateField):
    def __init__(self, *, max_value=None, min_value=None, **kwargs):
        self.max_value, self.min_value = max_value, min_value
        super().__init__(**kwargs)


class TimeField(TimeField):
    def __init__(self, *, max_value=None, min_value=None, **kwargs):
        self.max_value, self.min_value = max_value, min_value
        super().__init__(**kwargs)


class DateTimeField(DateTimeField, DateField):
    pass


__all__ = [
    'AutoField', 'BLANK_CHOICE_DASH', 'BigAutoField', 'BigIntegerField',
    'BinaryField', 'BooleanField', 'CharField', 'CommaSeparatedIntegerField',
    'DateField', 'DateTimeField', 'DecimalField', 'DurationField',
    'EmailField', 'Empty', 'Field', 'FieldDoesNotExist', 'FilePathField',
    'FloatField', 'GenericIPAddressField', 'IPAddressField', 'IntegerField',
    'NOT_PROVIDED', 'NullBooleanField', 'PositiveIntegerField',
    'PositiveSmallIntegerField', 'SlugField', 'SmallIntegerField', 'TextField',
    'TimeField', 'URLField', 'UUIDField',
]



from enum import Enum


class CEnum(Enum):
    @classmethod
    def list(cls):
        return [(tag.name, tag.value) for tag in cls]


class CREnum(Enum):
    @classmethod
    def list(cls):
        return [(tag.value, tag.name) for tag in cls]



import json
from functools import wraps

from django.http import HttpRequest
from smartify import Analyse as BAnalyse, PDict, P

from .classify import Classify
from .error import E
from .http_code import HttpCode as Hc


@E.register()
class AnalyseError:
    AE_METHOD_NOT_MATCH = E("Wrong request method", hc=Hc.InternalServerError)
    AE_REQUEST_NOT_FOUND = E("Can't find request", hc=Hc.InternalServerError)


class Analyse(BAnalyse):
    @classmethod
    def r(cls, b=None, q=None, a=None, method=None):
        """
        decorator for validating HttpRequest
        :param b: P list in it's BODY, in json format, without method in GET/DELETE
        :param q: P list in it's query
        :param a: P list in method/function argument
        :param method: Specify request method
        """
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                r = None
                for v in args:
                    if isinstance(v, HttpRequest):
                        r = v
                        break
                if not r:
                    for k in kwargs:
                        if isinstance(kwargs[k], HttpRequest):
                            r = kwargs[k]
                            break
                if not r:
                    raise AnalyseError.AE_REQUEST_NOT_FOUND
                if method and method != r.method:
                    raise AnalyseError.AE_METHOD_NOT_MATCH
                param_jar = dict()

                a_dict = kwargs or {}
                q_dict = r.GET.dict() or {}
                try:
                    b_dict = json.loads(r.body.decode())
                except json.JSONDecodeError:
                    b_dict = {}

                checks = [(a, a_dict), (b, b_dict), (q, q_dict)]
                for check in checks:
                    if check[0]:
                        if isinstance(check[0], P):
                            p = check[0]
                        else:
                            p = PDict().set_fields(*check[0])
                        _, result = p.run(check[1])
                        param_jar.update(result or {})

                r.d = Classify(param_jar)
                return func(r)

            return wrapper

        return decorator



from smartify import P, Attribute, BaseError, PDict, PList, Processor, PError, Symbol

from .excp import Excp
from .error import E
from .analyse import Analyse, AnalyseError
from .http_code import HttpCode
from .net_packer import NetPacker
from .models.base import ModelError

Hc = HttpCode

__all__ = [
    Excp, NetPacker,
    E, P, PList, PDict, PError, Processor, Attribute, BaseError, Symbol,
    Analyse, AnalyseError, HttpCode, Hc,
    ModelError,
]



import warnings
from functools import wraps

from .net_packer import NetPacker
from .middleware import HttpPackMiddleware
from .error import E


class Excp:
    @staticmethod
    def pack(func):
        warnings.warn(
            'Excp.pack is deprecated, use "raise XXError.xx" instead of "return XXError.xx".',
            DeprecationWarning)

        @wraps(func)
        def wrapper(*args, **kwargs):
            e = func(*args, **kwargs)
            if isinstance(e, E):
                raise e
            return e
        return wrapper

    # handle = HttpPackMiddleware
    @staticmethod
    def handle(get_response):
        warnings.warn(
            'Excp.handle is deprecated, use NetPacker.pack instead.',
            DeprecationWarning)
        return HttpPackMiddleware(get_response)

    @classmethod
    def http_response(cls, o, using_data_packer=True):
        warnings.warn(
            'Excp.http_response is deprecated, use NetPacker.send instead.',
            DeprecationWarning)
        return NetPacker.send(o, using_data_packer=using_data_packer)

    @classmethod
    def custom_http_response(cls, http_code_always=None, data_packer=None):
        warnings.warn(
            'Excp.custom_http_response is deprecated, use NetPacker.customize instead.',
            DeprecationWarning)
        NetPacker.customize(http_code_always, data_packer)

    @classmethod
    def debugging(cls, off=False):
        warnings.warn(
            'Excp.debugging is deprecated, use NetPacker.set_mode instead.',
            DeprecationWarning)
        NetPacker.set_mode(debug=not off)



from django.http import HttpResponse
from smartify import E


class HttpPackMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, r, *args, **kwargs):
        try:
            e = self.get_response(r, *args, **kwargs)
            if isinstance(e, HttpResponse):
                if e.content.decode().find(
                        "t return an HttpResponse object. It returned None instead.") == -1:
                    return e
                e = None
            if isinstance(e, E):
                raise e
        except E as err:
            e = err

        from .net_packer import NetPacker
        return NetPacker.send(e)

    def process_exception(self, _, e):
        from .net_packer import NetPacker
        if isinstance(e, E):
            return NetPacker.send(e)
        else:
            return None



class HttpCode:
    Continue = 100
    SwitchingProtocols = 101
    Processing = 102

    OK = 200
    Created = 201
    Accepted = 202
    NonAuthoritativeInformation = 203
    NoContent = 204
    ResetContent = 205
    PartialContent = 206
    MultiStatus = 207

    MultipleChoices = 300
    MovedPermanently = 301
    MoveTemporarily = 302
    SeeOther = 303
    NotModified = 304
    UseProxy = 305
    SwitchProxy = 306  # 不再使用
    TemporaryRedirect = 307

    BadRequest = 400
    Unauthorized = 401
    PaymentRequired = 402
    Forbidden = 403
    NotFound = 404
    MethodNotAllowed = 405
    NotAcceptable = 406
    ProxyAuthenticationRequired = 407
    RequestTimeout = 408
    Conflict = 409
    Gone = 410
    LengthRequired = 411
    PreconditionFailed = 412
    RequestEntityTooLarge = 413
    RequestURITooLong = 414
    UnsupportedMediaType = 415
    RequestedRangeNotSatisfiable = 416
    ExpectationFailed = 417
    ImATeapot = 418
    TooManyConnections = 421
    UnprocessableEntity = 422
    Locked = 423
    FailedDependency = 424
    TooEarly = 425
    UpgradeRequired = 426
    RetryWith = 449
    UnavailableForLegalReasons = 451

    InternalServerError = 500
    NotImplemented = 501
    BadGateway = 502
    ServiceUnavailable = 503
    GatewayTimeout = 504
    HTTPVersionNotSupported = 505
    VariantAlsoNegotiates = 506
    InsufficientStorage = 507
    BandwidthLimitExceeded = 509
    NotExtended = 510

    UnparseableResponseHeaders = 600



import gettext


class Translator:
    MODE_UPDATE = 0
    MODE_RELEASE = 1

    def __init__(self):
        self.langs = ['cn', 'en']
        self.localedir = './trans'
        self._translators = dict()
        self.current_lang = 'en'
        self.mode = self.MODE_UPDATE

    def load(self):
        self._translators = dict(zip(self.langs,
                                     gettext.translation('st', self.localedir, self.langs)))

    def set_lang(self, lang):
        if lang in self.langs:
            self.current_lang = lang
            return True
        return False

    def set_localedir(self, localedir):
        self.localedir = localedir


translator = Translator()
translator.load()



from django.db import models
from django.db.models import Q


class QuerySet(models.QuerySet):
    def search(self, *args, **kwargs):
        objects = self.all()
        cls = self.model

        field_dict = dict()
        for o_field in cls._meta.fields:
            field_dict[o_field.name] = o_field

        for q in args:
            if isinstance(q, Q):
                objects = objects.filter(q)

        for attr in kwargs:
            attr_value = kwargs[attr]

            if attr.endswith('__null'):
                attr = attr[:-6]
            elif attr_value is None:
                continue

            full = attr.endswith('__full')
            if full:
                attr = attr[:-6]

            filter_func = getattr(cls, '_search_%s' % attr, None)
            if filter_func and callable(filter_func):
                o_filter = filter_func(attr_value)
                if isinstance(o_filter, dict):
                    objects = objects.filter(**o_filter)
                elif isinstance(o_filter, Q):
                    objects = objects.filter(o_filter)
                continue

            if attr in field_dict:
                attr_field = field_dict[attr]
            else:
                attr_field = None

            filter_dict = dict()
            if not full and \
                    (isinstance(attr_field, models.CharField) or
                     isinstance(attr_field, models.TextField)):
                filter_dict.setdefault('%s__contains' % attr, attr_value)
            else:
                filter_dict.setdefault(attr, attr_value)

            objects = objects.filter(**filter_dict)
        return objects

    def dict(self, dictor):
        if callable(dictor):
            return list(map(dictor, self))
        return self

    def page(self, pager, last=0, count=5):
        return pager.page(self, last=last, count=count)


from .query import QuerySet


class Page:
    def __init__(self, object_list, next_value, total_count):
        self.object_list = object_list
        self.next_value = next_value
        self.total_count = total_count

    def dict(self, object_dictor=None, next_dictor=None):
        if callable(object_dictor):
            object_list = list(map(object_dictor, self.object_list))
        else:
            object_list = self.object_list
        if callable(next_dictor):
            next_value = next_dictor(self.next_value)
        else:
            next_value = self.next_value
        return dict(
            object_list=object_list,
            next_value=next_value,
            total_count=self.total_count,
        )


class Pager:
    FILTER_FIRST = 1
    CHOOSE_AMONG = 2

    def __init__(self, mode=FILTER_FIRST, compare_field='pk', ascend=True, order_by=None):
        self.mode = mode
        if self.mode == self.FILTER_FIRST:
            self.compared_field = compare_field
            self.filter_key = '%s__%st' % (compare_field, 'lg'[ascend])
            self.order_by = compare_field
            if not ascend:
                self.order_by = '-' + compare_field
            self.order_by = (self.order_by, )
        else:
            self.order_by = order_by

    def page(self, queryset: QuerySet, last=0, count=5):
        total_count = queryset.count()
        if self.mode == self.FILTER_FIRST:
            objects = queryset.filter(**{self.filter_key: last}).order_by(*self.order_by)
            page_count = objects.count()
            if page_count > count:
                objects = objects[:count]
                _next = getattr(objects[count-1], self.compared_field)
            else:
                _next = None
        else:
            objects = queryset.order_by(*self.order_by)
            objects = objects[last:]
            page_count = objects.count()
            if page_count > count:
                objects = objects[:count]
                _next = last + count
            else:
                _next = None
        return Page(objects, _next, total_count)



from .http_code import HttpCode as Hc
from smartify import E as BE, Attribute, BaseError, PError


class E(BE):
    def __init__(self, template: str, hc=Hc.OK):
        super(E, self).__init__(template)
        self.hc = hc

    def d(self):
        dict_ = super(E, self).d()
        dict_.update(Attribute.dictify(self, 'hc->http_code', 'eid->code'))
        del dict_['eid']
        return dict_

    def d_debug(self):
        return super(E, self).d_debug()


def error_update(e: BE, hc=Hc.OK):
    e.__class__ = E
    setattr(e, 'hc', hc)


error_update(BaseError.OK, hc=Hc.OK)
error_update(BaseError.ERROR_GENERATE, hc=Hc.InternalServerError)
error_update(PError.VALIDATOR_CRUSHED, hc=Hc.InternalServerError)
error_update(PError.PROCESSOR_CRUSHED, hc=Hc.InternalServerError)
error_update(PError.REQUIRE_DICT, hc=Hc.Forbidden)
error_update(PError.REQUIRE_LIST, hc=Hc.Forbidden)
error_update(PError.NULL_NOT_ALLOW, hc=Hc.Forbidden)
error_update(PError.PATTERN_NOT_MATCH, hc=Hc.Forbidden)



from setuptools import setup, find_packages

setup(
    name='SmartDjango',
    version='3.6.5 alpha3',
    keywords=('django',),
    description='更高效率的Django开发[Chinese Version]',
    long_description='提供智能模型用于字段检测，数据检索，错误类等',
    license='MIT Licence',
    url='https://github.com/lqj679ssn/SmartDjango',
    author='Adel Liu',
    author_email='i@6-79.cn',
    platforms='any',
    packages=find_packages(),
    install_requires=[
        'django>=2.2.5',
        'smartify>=0.0.2',
    ],
)



import json

from .middleware import HttpPackMiddleware
from django.http import HttpResponse
from smartify import BaseError

from .error import E
from .http_code import HttpCode as Hc


@E.register()
class PackerError:
    HTTP_DATA_PACKER = E("Http data packer crashed", hc=Hc.InternalServerError)


class NetPacker:
    fixed_code = False
    debug = True
    data_packer = None
    msg_as_headers = False

    @staticmethod
    def pack(get_response):
        return HttpPackMiddleware(get_response)

    @classmethod
    def send(cls, o, using_data_packer=True, headers=None):
        body, e = (None, o) if isinstance(o, E) else (o, BaseError.OK())

        headers = headers or dict()
        resp = e.d_debug() if cls.debug else e.d()
        if cls.msg_as_headers:
            headers.update(resp)
            resp = body
        else:
            resp['body'] = body

        if using_data_packer and cls.data_packer:
            try:
                resp = cls.data_packer(resp)
            except Exception as err:
                return cls.send(
                    PackerError.HTTP_DATA_PACKER(debug_message=err), using_data_packer=False)

        response = HttpResponse(
            json.dumps(resp, ensure_ascii=False),
            status=cls.fixed_code or e.hc,
            content_type="application/json; encoding=utf-8",
        )
        for key in headers:
            response[key] = headers[key]
        return response

    @classmethod
    def customize(cls, fixed_http_code=None, data_packer=None):
        cls.fixed_code = int(fixed_http_code)
        cls.data_packer = data_packer if callable(data_packer) else None

    @classmethod
    def set_mode(cls, debug=False):
        cls.debug = debug
